@@ -169,6 +169,10 @@ static int parse_auth_choice(struct openconnect_info *vpninfo, struct oc_auth_fo
 		choice->override_label = (char *)xmlGetProp(xml_node, (unsigned char *)"override-label");
 	}
 
+	if (!strcmp(opt->form.name, "group_list")) {
+		form->authgroup_opt = opt;
+	}
+
 	/* We link the choice _first_ so it's at the top of what we present
 	   to the user */
 	opt->form.next = form->opts;
@@ -176,11 +180,6 @@ static int parse_auth_choice(struct openconnect_info *vpninfo, struct oc_auth_fo
 	return 0;
 }
 
-/* Return value:
- *  < 0, on error
- *  = 0, when form was cancelled
- *  = 1, when form was parsed
- */
 static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *form,
 		      xmlNode *xml_node)
 {
@@ -570,6 +569,14 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response, struct
 		goto out;
 	}
 
+	if (form->authgroup_opt && form->authgroup_opt->nr_choices > 1 && vpninfo->xmlpost) {
+		/* XML POST + authgroups require NEWGROUP support, so disable for now */
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Disabling XML POST due to authgroup option\n"));
+		ret = -EINVAL;
+		goto out;
+	}
+
 	*formp = form;
 	xmlFreeDoc(xml_doc);
 	return 0;
@@ -780,11 +787,17 @@ int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, in
 	xmlNodePtr root, node;
 	xmlDocPtr doc = xmlpost_new_query(vpninfo, "init", &root);
 	char *url;
+	int result;
 
 	if (!doc)
 		return -ENOMEM;
 
-	if (asprintf(&url, "https://%s", vpninfo->hostname) == -1)
+	if (vpninfo->urlpath)
+		result = asprintf(&url, "https://%s/%s", vpninfo->hostname, vpninfo->urlpath);
+	else
+		result = asprintf(&url, "https://%s", vpninfo->hostname);
+
+	if (result == -1)
 		goto bad;
 	node = xmlNewTextChild(root, NULL, XCAST("group-access"), XCAST(url));
 	free(url);
@@ -795,6 +808,11 @@ int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, in
 		if (!node)
 			goto bad;
 	}
+	if (vpninfo->authgroup) {
+		node = xmlNewTextChild(root, NULL, XCAST("group-select"), XCAST(vpninfo->authgroup));
+		if (!node)
+			goto bad;
+	}
 	return xmlpost_complete(doc, request_body, req_len);
 
 bad:
