@@ -517,7 +517,7 @@ static int fetch_config(struct openconnect_info *vpninfo, char *fu, char *bu,
 
 static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int buflen)
 {
-	char fname[16];
+	char fname[64];
 	int fd, ret;
 
 	if (!vpninfo->csd_wrapper && !buflen) {
@@ -541,7 +541,8 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 
 	fname[0] = 0;
 	if (buflen) {
-		sprintf(fname, "/tmp/csdXXXXXX");
+		char *tmpdir = getenv("TMPDIR");
+		snprintf(fname, 64, "%s/csdXXXXXX", tmpdir ? tmpdir : "/tmp");
 		fd = mkstemp(fname);
 		if (fd < 0) {
 			int err = -errno;
@@ -758,15 +759,10 @@ static int handle_redirect(struct openconnect_info *vpninfo)
 		}
 
 		if (strcasecmp(vpninfo->hostname, host) || port != vpninfo->port) {
-			free(vpninfo->unique_hostname);
-			vpninfo->unique_hostname = NULL;
-			free(vpninfo->hostname);
-			vpninfo->hostname = host;
+			openconnect_set_hostname(vpninfo, host);
 			vpninfo->port = port;
 
 			/* Kill the existing connection, and a new one will happen */
-			free(vpninfo->peer_addr);
-			vpninfo->peer_addr = NULL;
 			openconnect_close_https(vpninfo, 0);
 			clear_cookies(vpninfo);
 			vpninfo->redirect_type = REDIR_TYPE_NEWHOST;
@@ -920,6 +916,7 @@ static int do_https_request(struct openconnect_info *vpninfo, const char *method
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to open HTTPS connection to %s\n"),
 				     vpninfo->hostname);
+			buf_free(buf);
 			return -EINVAL;
 		}
 	}
