@@ -27,6 +27,8 @@
 #ifndef __OPENCONNECT_INTERNAL_H__
 #define __OPENCONNECT_INTERNAL_H__
 
+#define __OPENCONNECT_PRIVATE__
+
 #include "openconnect.h"
 
 #if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
@@ -71,7 +73,7 @@
 #include <libintl.h>
 #define _(s) dgettext("openconnect", s)
 #else
-#define _(s) s
+#define _(s) ((char *)(s))
 #endif
 #define N_(s) s
 
@@ -171,8 +173,6 @@ struct openconnect_info {
 	const char *servercert;
 	const char *xmlconfig;
 	char xmlsha1[(SHA1_SIZE * 2) + 1];
-	char *username;
-	char *password;
 	char *authgroup;
 	int nopasswd;
 	int xmlpost;
@@ -224,6 +224,7 @@ struct openconnect_info {
 #endif
 	gnutls_privkey_t my_pkey;
 	gnutls_x509_crt_t *my_certs;
+	uint8_t *free_my_certs;
 	unsigned int nr_my_certs;
 #endif
 #endif /* OPENCONNECT_GNUTLS */
