@@ -35,19 +35,23 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#if defined(__linux__) || defined(ANDROID)
+#if defined(__linux__) || defined(__ANDROID__)
 #include <sys/vfs.h>
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__OpenBSD__) || defined(__APPLE__)
 #include <sys/param.h>
 #include <sys/mount.h>
-#elif defined (__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
+#elif defined(__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
 #include <sys/statvfs.h>
-#elif defined (__GNU__)
+#elif defined(__GNU__)
 #include <sys/statfs.h>
 #endif
 
 #include "openconnect-internal.h"
 
+#ifdef ANDROID_KEYSTORE
+#include <sys/un.h>
+#endif
+
 /* OSX < 1.6 doesn't have AI_NUMERICSERV */
 #ifndef AI_NUMERICSERV
 #define AI_NUMERICSERV 0
@@ -74,7 +78,7 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 		if (vpninfo->cancel_fd > sockfd)
 			maxfd = vpninfo->cancel_fd;
 	}
-	
+
 	/* Later we'll render this whole exercise non-pointless by
 	   including a 'cancelfd' here too. */
 	select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
@@ -83,7 +87,7 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 		errno = EINTR;
 		return -1;
 	}
-		
+
 	/* Check whether connect() succeeded or failed by using
 	   getpeername(). See http://cr.yp.to/docs/connect.html */
 	return getpeername(sockfd, (void *)&peer, &peerlen);
@@ -111,17 +115,18 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 		if (cancellable_connect(vpninfo, ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen)) {
 		reconn_err:
 			if (vpninfo->proxy) {
-				vpn_progress(vpninfo, PRG_ERR, 
+				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to reconnect to proxy %s\n"),
 					     vpninfo->proxy);
 			} else {
-				vpn_progress(vpninfo, PRG_ERR, 
+				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to reconnect to host %s\n"),
 					     vpninfo->hostname);
 			}
+			if (ssl_sock >= 0)
+				close(ssl_sock);
 			return -EINVAL;
 		}
-		
 	} else {
 		struct addrinfo hints, *result, *rp;
 		char *hostname;
@@ -204,28 +209,32 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 		}
 
 		err = getaddrinfo(hostname, port, &hints, &result);
-		if (hints.ai_flags & AI_NUMERICHOST)
-			free(hostname);
 
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("getaddrinfo failed for host '%s': %s\n"),
 				     hostname, gai_strerror(err));
+			if (hints.ai_flags & AI_NUMERICHOST)
+				free(hostname);
 			return -EINVAL;
 		}
+		if (hints.ai_flags & AI_NUMERICHOST)
+			free(hostname);
 
 		for (rp = result; rp ; rp = rp->ai_next) {
 			char host[80];
 
+			host[0] = 0;
 			if (!getnameinfo(rp->ai_addr, rp->ai_addrlen, host,
 					 sizeof(host), NULL, 0, NI_NUMERICHOST))
-				vpn_progress(vpninfo, PRG_INFO,
-					     _("Attempting to connect to %s%s%s:%s\n"),
-					     rp->ai_family == AF_INET6?"[":"",
+				vpn_progress(vpninfo, PRG_INFO, vpninfo->proxy_type ?
+						     _("Attempting to connect to proxy %s%s%s:%s\n") :
+						     _("Attempting to connect to server %s%s%s:%s\n"),
+					     rp->ai_family == AF_INET6 ? "[" : "",
 					     host,
-					     rp->ai_family == AF_INET6?"]":"",
+					     rp->ai_family == AF_INET6 ? "]" : "",
 					     port);
-			
+
 			ssl_sock = socket(rp->ai_family, rp->ai_socktype,
 					  rp->ai_protocol);
 			if (ssl_sock < 0)
@@ -242,13 +251,30 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				}
 				vpninfo->peer_addrlen = rp->ai_addrlen;
 				memcpy(vpninfo->peer_addr, rp->ai_addr, rp->ai_addrlen);
+				/* If no proxy, and if more than one address for the hostname,
+				   ensure that we output the same IP address in authentication
+				   results (from libopenconnect or --authenticate). */
+				if (!vpninfo->proxy && (rp != result || rp->ai_next) && host[0]) {
+					char *p = malloc(strlen(host) + 3);
+					if (p) {
+						free(vpninfo->unique_hostname);
+						vpninfo->unique_hostname = p;
+						if (rp->ai_family == AF_INET6)
+							*p++ = '[';
+						memcpy(p, host, strlen(host));
+						p += strlen(host);
+						if (rp->ai_family == AF_INET6)
+							*p++ = ']';
+						*p = 0;
+					}
+				}
 				break;
 			}
 			close(ssl_sock);
 			ssl_sock = -1;
 		}
 		freeaddrinfo(result);
-		
+
 		if (ssl_sock < 0) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to connect to host %s\n"),
@@ -355,7 +381,7 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 }
 #endif
 
-#if defined(OPENCONNECT_OPENSSL) || defined (DTLS_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
 /* We put this here rather than in openssl.c because it might be needed
    for OpenSSL DTLS support even when GnuTLS is being used for HTTPS */
 int openconnect_print_err_cb(const char *str, size_t len, void *ptr)
@@ -408,14 +434,24 @@ int keystore_fetch(const char *key, unsigned char **result)
 	close(fd);
 	return ret;
 }
-#elif defined (ANDROID_KEYSTORE)
-#include <cutils/sockets.h>
-#include <keystore.h>
-char *keystore_strerror(int err)
+#elif defined(ANDROID_KEYSTORE)
+/* keystore.h isn't in the NDK so we need to define these */
+#define NO_ERROR		1
+#define LOCKED			2
+#define UNINITIALIZED		3
+#define SYSTEM_ERROR		4
+#define PROTOCOL_ERROR		5
+#define PERMISSION_DENIED	6
+#define KEY_NOT_FOUND		7
+#define VALUE_CORRUPTED		8
+#define UNDEFINED_ACTION	9
+#define WRONG_PASSWORD		10
+
+const char *keystore_strerror(int err)
 {
 	switch (-err) {
 	case NO_ERROR:		return _("No error");
-	case LOCKED:		return _("Keystore ocked");
+	case LOCKED:		return _("Keystore locked");
 	case UNINITIALIZED:	return _("Keystore uninitialized");
 	case SYSTEM_ERROR:	return _("System error");
 	case PROTOCOL_ERROR:	return _("Protocol error");
@@ -423,10 +459,10 @@ char *keystore_strerror(int err)
 	case KEY_NOT_FOUND:	return _("Key not found");
 	case VALUE_CORRUPTED:	return _("Value corrupted");
 	case UNDEFINED_ACTION:	return _("Undefined action");
-	case WRONG_PASSWORD_0:
-	case WRONG_PASSWORD_1:
-	case WRONG_PASSWORD_2:
-	case WRONG_PASSWORD_3:	return _("Wrong password");
+	case WRONG_PASSWORD:
+	case WRONG_PASSWORD+1:
+	case WRONG_PASSWORD+2:
+	case WRONG_PASSWORD+3:	return _("Wrong password");
 	default:		return _("Unknown error");
 	}
 }
@@ -435,17 +471,21 @@ char *keystore_strerror(int err)
    own strerror function above). The numbers are from Android's keystore.h */
 int keystore_fetch(const char *key, unsigned char **result)
 {
+	struct sockaddr_un sa = { AF_UNIX, "/dev/socket/keystore" };
+	socklen_t sl = offsetof(struct sockaddr_un, sun_path) + strlen(sa.sun_path) + 1;
 	unsigned char *data, *p;
 	unsigned char buf[3];
-	int len, fd, ofs;
+	int len, fd;
 	int ret = -SYSTEM_ERROR;
 
-	fd = socket_local_client("keystore",
-				 ANDROID_SOCKET_NAMESPACE_RESERVED,
-				 SOCK_STREAM);
+	fd = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (fd < 0)
 		return -SYSTEM_ERROR;
 
+	if (connect(fd, (void *)&sa, sl)) {
+		close(fd);
+		return -SYSTEM_ERROR;
+	}
 	len = strlen(key);
 	buf[0] = 'g';
 	buf[1] = len >> 8;
