@@ -53,7 +53,7 @@
 #include "openconnect-internal.h"
 
 /*
- * If an if_tun.h include file was found anywhere (by the Makefile), it's 
+ * If an if_tun.h include file was found anywhere (by the Makefile), it's
  * included. Else, we end up assuming that we have BSD-style devices such
  * as /dev/tun0 etc.
  */
@@ -126,7 +126,7 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 				 int *v6_incs)
 {
 	struct in_addr addr;
-	const char *in_ex = include?"IN":"EX";
+	const char *in_ex = include ? "IN" : "EX";
 	char envname[80];
 	char *slash;
 
@@ -158,7 +158,7 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 		(*v6_incs)++;
 		return 0;
 	}
-		
+
 	if (!inet_aton(route, &addr)) {
 		*slash = '/';
 		goto badinc;
@@ -233,20 +233,20 @@ static void set_banner(struct openconnect_info *vpninfo)
 	}
 	p = vpninfo->banner;
 	q = banner;
-	
+
 	while (*p) {
 		if (*p == '%' && isxdigit((int)(unsigned char)p[1]) &&
 		    isxdigit((int)(unsigned char)p[2])) {
 			*(q++) = unhex(p + 1);
 			p += 3;
-		} else 
+		} else
 			*(q++) = *(p++);
 	}
 	*q = 0;
 	setenv("CISCO_BANNER", banner, 1);
 
 	free(banner);
-}	
+}
 
 static void set_script_env(struct openconnect_info *vpninfo)
 {
@@ -306,7 +306,8 @@ static void set_script_env(struct openconnect_info *vpninfo)
 
 	if (vpninfo->vpn_domain)
 		setenv("CISCO_DEF_DOMAIN", vpninfo->vpn_domain, 1);
-	else unsetenv ("CISCO_DEF_DOMAIN");
+	else
+		unsetenv("CISCO_DEF_DOMAIN");
 
 	if (vpninfo->vpn_proxy_pac)
 		setenv("CISCO_PROXY_PAC", vpninfo->vpn_proxy_pac, 1);
@@ -376,7 +377,7 @@ int script_config_tun(struct openconnect_info *vpninfo, const char *reason)
 {
 	int ret;
 
-	if (!vpninfo->vpnc_script)
+	if (!vpninfo->vpnc_script || vpninfo->script_tun)
 		return 0;
 
 	setenv("reason", reason, 1);
@@ -412,7 +413,7 @@ static int link_proto(int unit_nr, const char *devname, uint64_t flags)
 
 	tun2_fd = open("/dev/tun", O_RDWR);
 	if (tun2_fd < 0) {
-		perror(_("Could not /dev/tun for plumbing"));
+		perror(_("Could not open /dev/tun for plumbing"));
 		return -EIO;
 	}
 	if (ioctl(tun2_fd, I_PUSH, "ip") < 0) {
@@ -523,7 +524,7 @@ static int os_setup_tun(struct openconnect_info *vpninfo)
 	}
 	if (!vpninfo->ifname)
 		vpninfo->ifname = strdup(ifr.ifr_name);
-#elif defined (__sun__)
+#elif defined(__sun__)
 	static char tun_name[80];
 	int unit_nr;
 
@@ -652,6 +653,8 @@ int setup_tun(struct openconnect_info *vpninfo)
 			perror(_("fork"));
 			exit(1);
 		} else if (!child) {
+			if (setpgid(0, getpid()) < 0)
+				perror(_("setpgid"));
 			close(tun_fd);
 			setenv_int("VPNFD", fds[1]);
 			execl("/bin/sh", "/bin/sh", "-c", vpninfo->vpnc_script, NULL);
@@ -784,9 +787,10 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 }
 
 void shutdown_tun(struct openconnect_info *vpninfo)
-{	
+{
 	if (vpninfo->script_tun) {
-		kill(vpninfo->script_tun, SIGHUP);
+		/* nuke the whole process group */
+		kill(-vpninfo->script_tun, SIGHUP);
 	} else {
 		script_config_tun(vpninfo, "disconnect");
 #ifdef __sun__
